from typing import Iterator
from typing import List
from typing import Tuple


def list_to_ranges(lines: List[int]) -> List[Tuple[int, int]]:
    """Produce a list of ranges for `format_lines`.

    >>> list_to_ranges([1,2,3,4])
    [(1, 4)]
    >>> list_to_ranges([1,4])
    [(1, 1), (4, 4)]
    >>> list_to_ranges([1,2,4,5,7])
    [(1, 2), (4, 5), (7, 7)]
    >>> list_to_ranges([-2, -1, 0, 1])
    [(-2, 1)]
    >>> list_to_ranges([])
    []


    """
    lines = sorted(lines)

    pairs = []
    start = None
    prev_line = None
    for line in lines:
        if start is None:
            start = prev_line = line
            continue

        if line == prev_line + 1:
            prev_line = line
            continue

        pairs.append((start, prev_line))
        start = prev_line = line

    if start and prev_line:
        pairs.append((start, prev_line))

    return pairs


def format_ranges(pairs: List[Tuple[int, int]]) -> Iterator[str]:
    """
    >>> list(format_ranges([(1,2), (3,3), (4,20)]))
    ['1-2', '3', '4-20']
    """
    for (start, end) in pairs:
        if start == end:
            yield f"{start}"
        else:
            yield f"{start}-{end}"


def format_lines(lines: List[int]):
    return ", ".join(format_ranges(list_to_ranges(lines)))



import pathlib

import pytest
import typer

from .cov import mark_coveraged_lines_per_file
from .cov import mark_executable_lines_per_file
from .format import format_lines
from .git import get_new_lines_per_file


def make_relative(path: pathlib.Path) -> str:
    return str(path.relative_to(str(pathlib.Path(".").resolve())))


def nl_cov(revision: str = "master") -> float:
    """
    Get the coverage for new lines only.
    """
    typer.echo("New Line Coverage")
    typer.echo("")

    new_lines_per_file = get_new_lines_per_file(revision, glob="*.py")

    mark_executable_lines_per_file(new_lines_per_file)
    mark_coveraged_lines_per_file(new_lines_per_file)

    # Prepare the formatting strings, header, and column sorting.
    max_name = max([len(make_relative(p)) for p in new_lines_per_file] + [5])
    fmt_name = "%%- %ds  " % max_name
    fmt_skip_covered = "\n%s file%s skipped due to complete coverage."
    fmt_skip_empty = "\n%s empty file%s skipped."

    header = (fmt_name % "Name") + " Lines   Miss"
    fmt_coverage = fmt_name + "%6d %6d"

    header += "%*s" % (9, "Cover")
    fmt_coverage += " %%%ds" % (8,)

    header += "   Missing"
    fmt_coverage += "   %s"

    rule = "-" * len(header)

    # Write the header
    typer.echo(header)
    typer.echo(rule)

    total_num_newlines = 0
    total_num_covered = 0
    for (p, new_lines) in new_lines_per_file.items():
        num_newlines = len(
            [lineno for lineno, line in new_lines.items() if line.is_empty is False and line.is_executable is True]
        )

        if num_newlines == 0:
            continue

        num_covered = len(
            [
                lineno
                for lineno, line in new_lines.items()
                if line.is_executed is True and line.is_empty is False and line.is_executable is True
            ]
        )
        total_num_newlines += num_newlines
        total_num_covered += num_covered

        num_uncovered = num_newlines - num_covered

        args = (
            make_relative(p),
            num_newlines,
            num_uncovered,
            f"{num_covered / num_newlines:.0%}",
            format_lines(
                [
                    lineno
                    for lineno, line in new_lines.items()
                    if line.is_executed is not True and line.is_empty is False and line.is_executable is True
                ]
            ),
        )

        text = fmt_coverage % args

        typer.echo(text)

    # Write a TOTAL line if we had at least one file.
    if total_num_newlines > 0:
        typer.echo(rule)
        args = (
            "TOTAL",
            total_num_newlines,
            total_num_newlines - total_num_covered,
            f"{total_num_covered / total_num_newlines:.0%}",
            "",
        )

        typer.echo(fmt_coverage % args)

    return total_num_covered / total_num_newlines if total_num_newlines else 1


def cli():
    typer.run(nl_cov)


def validate_fail_under(num_str):
    try:
        return int(num_str)
    except ValueError:
        return float(num_str)


def pytest_addoption(parser):
    group = parser.getgroup("coverage for new lines")
    group.addoption(
        "--nlcov-revision",
        action="store",
        default="master",
        dest="nlcov_revision",
        help="Revision to determine the added lines",
    )
    group.addoption(
        "--nlcov-fail-under",
        action="store",
        type=validate_fail_under,
        help="Fail if the total coverage is less",
    )


class NLCovPlugin:
    @pytest.mark.trylast
    def pytest_terminal_summary(self, terminalreporter, config):
        cov_plugin = config.pluginmanager.get_plugin("_cov")
        if cov_plugin.cov_controller is None:
            terminalreporter.write_line("No _cov plugin available")
            return

        if cov_plugin.cov_total is None:
            terminalreporter.write_line("cov_total is None")
            return

        cov = cov_plugin.cov_controller.cov
        if cov is None:
            terminalreporter.write_line("Cov is None")
            return

        coverage = nl_cov(config.option.nlcov_revision)

        if config.option.cov_fail_under is not None and config.option.cov_fail_under > 0:
            failed = coverage < config.option.cov_fail_under
            terminalreporter.write(
                f'{"FAIL " if failed else ""}Required test coverage of {config.option.cov_fail_under}% {"not reached" if failed else "reached"}. '
                f"Total coverage: {coverage:.2f}%\n",
                **({"red": True, "bold": True} if failed else {"green": True}),
            )


def pytest_configure(config):  # pragma: no cover
    if not config.pluginmanager.has_plugin("_cov"):
        typer.echo("nlcov is installed, but pytest-cov is not installed, so nlcov will not be executed.")
        return
    config.pluginmanager.register(NLCovPlugin())



import pathlib
from io import StringIO
from typing import Optional

import git
from unidiff import PatchSet

from .data import LinesPerFile
from .data import lines_per_file_factory


def get_diff(revision: str):
    return git.Repo(".").git.diff(revision, ignore_blank_lines=True, ignore_space_at_eol=True)


def get_new_lines_per_file(revision: str, glob: Optional[str] = None) -> LinesPerFile:
    """
    Get all lines which are new according to git.
    """
    uni_diff_text = get_diff(revision)

    lines_per_file = lines_per_file_factory()

    for patched_file in PatchSet(StringIO(uni_diff_text)):
        file_path = pathlib.Path(patched_file.path).resolve()
        if glob and not file_path.match(glob):
            continue
        for hunk in patched_file:
            for hunk_line in hunk:
                if hunk_line.target_line_no is None or not hunk_line.is_added:
                    continue
                line = lines_per_file[file_path][hunk_line.target_line_no]
                line.is_empty = hunk_line.value.strip() == ""

    return lines_per_file



import pathlib

from coverage import CoverageData
from coverage import parser

from .data import LinesPerFile


def get_coverage_data():
    coverage_data = CoverageData()
    coverage_data.read()

    return {
        pathlib.Path(measured_file).resolve(): coverage_data.lines(measured_file)
        for measured_file in coverage_data.measured_files()
    }


def mark_coveraged_lines_per_file(lines_per_file: LinesPerFile):
    """ """
    coverage_data = get_coverage_data()

    for measured_file, coveraged_linenos in coverage_data.items():
        if measured_file not in lines_per_file:
            continue
        for coveraged_lineno in coveraged_linenos:
            if coveraged_lineno not in lines_per_file[measured_file]:
                continue
            lines_per_file[measured_file][coveraged_lineno].is_executed = True
            for subline in lines_per_file[measured_file][coveraged_lineno].multilines or []:
                lines_per_file[measured_file][subline].is_executed = True


def get_python_parser(file_path) -> parser.PythonParser:
    parsed_file = parser.PythonParser(filename=file_path)
    parsed_file.parse_source()

    return parsed_file


def mark_executable_lines_per_file(lines_per_file: LinesPerFile) -> LinesPerFile:
    """
    Mark the executable lines based on a Python parser.
    When a line is marked as new and it belongs to a multiline statement,
    the first line of that statement is added and marked as new too.
    """
    for file_path, lines in lines_per_file.items():
        parsed_file = get_python_parser(file_path)
        executable_linenos = parsed_file.statements
        for lineno, line in list(lines.items()):
            if lineno in executable_linenos:
                line.is_executable = True
            lineno_of_multiline_statement = parsed_file.first_line(lineno)
            if lineno_of_multiline_statement != lineno:
                lines[lineno].is_executable = lines[lineno_of_multiline_statement].is_executable = True
                multilines = lines[lineno_of_multiline_statement].multilines or []
                multilines.append(lineno)
                lines[lineno_of_multiline_statement].multilines = multilines



import collections
import dataclasses
import pathlib
from typing import Dict
from typing import List
from typing import Optional


@dataclasses.dataclass
class Line:
    is_empty: Optional[bool] = None
    is_executed: Optional[bool] = None
    is_executable: Optional[bool] = None
    multilines: Optional[List[int]] = None  # all lines which belong to this starting line of a multiline statement


Lines = Dict[int, Line]
LinesPerFile = Dict[pathlib.Path, Lines]


def lines_per_file_factory():
    return collections.defaultdict(lambda: collections.defaultdict(Line))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pytest_nlcov']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.17,<4.0.0',
 'coverage>=5.5,<6.0',
 'pytest-cov>=2.10.1',
 'typer>=0.3.2,<0.4.0',
 'unidiff>=0.6.0,<0.7.0']

entry_points = \
{'console_scripts': ['nlcov = pytest_nlcov:cli'],
 'pytest11': ['nlcov = pytest_nlcov']}

setup_kwargs = {
    'name': 'pytest-nlcov',
    'version': '0.1.0',
    'description': '',
    'long_description': '# pytest_nlcov\n\nWith `pytest_nlcov` you can check the test coverage of new lines only. It will\ncheck git for added and modified lines and will compute the coverage\njust for those lines\n\n## Installation\n\n```sh\npip install pytest_nlcov\n```\n\nNote: `pytest_cov` is required and will be automatically installed when it \nis not installed yet.\n\n## Usage with pytest\n\nWhen `pytest_nlcov` is installed, it will be discovered by pytest and executed as last step to\nshow you the test coverage of new lines.\n\n```sh\npytest\n```\n\nTwo option can be given:\n\n- revision\n- fail threshold\n\n### Revision\n\nDefault, the new lines are based on the git diff with master. You can specify other revisions.\n\n```sh\npytest --nlcov-revision main\n```\n\n### Fail Threshold\n\nOptionally you can add a threshold to fail the tests when the coverage is below the threshold.\n\n```sh\npytest --nlcov-fail-under 0.6\n```\n\n## Usage without pytest\n\n`pytest_nlcov` can be run without pytest. Therefor you have to run `coverage` first, because `pytest_nlcov`\nneeds its coverage data.\n\n```sh\ncoverage\nnlcov\n```\n\nOptionally a revision can be given\n\n```sh\nnlcov main\n```\n',
    'author': 'Marc Rijken',
    'author_email': 'marc@rijken.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mrijken/pytest_nlcov',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
